import googleSheets from '@googleapis/sheets';
import { b as blue, u as underline, C as ColumnType } from './cli.js';
import tinydate from 'tinydate';
import http from 'node:http';
import { OAuth2Client } from 'google-auth-library';
import open from 'open';
import 'node:module';
import 'node:os';
import 'node:path';
import 'log-update';
import 'sade';
import 'simplecrawler';
import 'globrex';
import 'node:url';
import 'node:worker_threads';
import 'node:child_process';
import 'chrome-launcher';
import 'node:fs/promises';

const CLIENT_ID = "172798687411-4459o482o4trfp3kvkkoer5m7qiiejpr.apps.googleusercontent.com";
const CLIENT_SECRET = "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";
const PORT = "2091";
const getAuthenticatedClient = async () => {
  const oAuth2Client = new OAuth2Client(CLIENT_ID, CLIENT_SECRET, `http://localhost:${PORT}/oauth2callback`);
  const authorizeUrl = oAuth2Client.generateAuthUrl({
    access_type: "offline",
    scope: ["https://www.googleapis.com/auth/spreadsheets"]
  });
  return new Promise((resolve, reject) => {
    const server = http.createServer(async (req, res) => {
      const url = req.url;
      try {
        if (url.includes("/oauth2callback")) {
          const qs = new URL(url, `http://localhost:${PORT}`).searchParams;
          const code = qs.get("code");
          console.log("Google authentication complete");
          res.setHeader("content-type", "text/html");
          res.write(`
<h1>Authentication successful! Creating spreadsheet...</h1>
<style>
body {
  text-align: center;
  font-family: system-ui, sans-serif;
  margin: 40px;
}
</style>
`);
          const { tokens } = await oAuth2Client.getToken(code);
          oAuth2Client.setCredentials(tokens);
          let didClose = false;
          resolve({
            auth: oAuth2Client,
            redirect: (url2) => {
              if (didClose)
                return;
              didClose = true;
              res.end(`<script>location = ${JSON.stringify(url2)}<\/script>`);
              server.close();
            }
          });
          setTimeout(() => {
            if (didClose)
              return;
            didClose = true;
            res.end();
            server.close();
          }, 1e4);
        }
      } catch (error) {
        reject(error);
      }
    }).listen(PORT, () => {
      console.log(`Waiting for google authentication: ${blue(underline(authorizeUrl))}`);
      open(authorizeUrl, { wait: false }).then((cp) => cp.unref());
    });
  });
};

const getCharts = (sheetId, auditBreakdownSheetId, columns) => {
  let chartRow = 0;
  const charts = [];
  const batchUpdates = columns.map((column, i) => {
    if (column.field.type === ColumnType.AuditScore && column.field.hasAuditValueColumn || column.field.type === ColumnType.AuditBoolean)
      return null;
    const thisChartRow = chartRow;
    chartRow += 15;
    const columnIndex = i + 1;
    charts.push({
      column,
      chartRow: thisChartRow,
      columnIndex
    });
    return {
      addChart: {
        chart: {
          position: {
            overlayPosition: {
              anchorCell: {
                sheetId: auditBreakdownSheetId,
                columnIndex: 0,
                rowIndex: thisChartRow
              },
              widthPixels: 800,
              heightPixels: 300
            }
          },
          spec: {
            title: `Distribution of ${column.name} (${column.nameDetail})`,
            histogramChart: {
              series: [
                {
                  data: {
                    sourceRange: {
                      sources: [
                        {
                          sheetId,
                          startRowIndex: 2,
                          startColumnIndex: columnIndex,
                          endColumnIndex: columnIndex + 1
                        }
                      ]
                    }
                  }
                }
              ]
            }
          }
        }
      }
    };
  }).filter(Boolean);
  const batchValues = {
    data: charts.map((chart) => {
      const columnLetter = numToSSColumn(chart.columnIndex + 1);
      const isScore = chart.column.field.type === ColumnType.AuditScore || chart.column.field.type === ColumnType.CategoryScore;
      const numRowsHighestOrLowestShown = 8;
      return {
        range: `'${sheetNames.auditBreakdown}'!J${chart.chartRow + 1}`,
        values: [
          [
            "Mean:",
            `=AVERAGE('${sheetNames.main}'!${columnLetter}$3:${columnLetter})`
          ],
          [
            "Median:",
            `=MEDIAN('${sheetNames.main}'!${columnLetter}$3:${columnLetter})`
          ],
          [
            "Highest:",
            `=MAX('${sheetNames.main}'!${columnLetter}$3:${columnLetter})`
          ],
          [
            "Lowest:",
            `=MIN('${sheetNames.main}'!${columnLetter}$3:${columnLetter})`
          ],
          [
            isScore ? "Lowest scores:" : `Highest values (${chart.column.nameDetail}):`
          ],
          [
            `=SORTN(
              {'${sheetNames.main}'!${columnLetter}$3:${columnLetter},'${sheetNames.main}'!$A$3:$A},
              ${numRowsHighestOrLowestShown},
              0,
              1,
              ${isScore}
            )`.replace(/\n\s*/g, "")
          ]
        ]
      };
    })
  };
  return { batchUpdates, batchValues };
};
const numToSSColumn = (num) => {
  let s = "";
  let t;
  while (num > 0) {
    t = (num - 1) % 26;
    s = String.fromCodePoint(65 + t) + s;
    num = Math.trunc((num - t) / 26);
  }
  return s || void 0;
};

const makeColor = (red, green, blue) => ({
  red: red / 255,
  green: green / 255,
  blue: blue / 255
});
const colors = {
  bad: makeColor(242, 112, 65),
  good: makeColor(113, 244, 174),
  neutral: makeColor(255, 255, 0)
};
const getInitialFormatting = (sheetId, runInfoSheetId) => [
  {
    updateDimensionProperties: {
      range: {
        sheetId,
        dimension: "COLUMNS",
        startIndex: 0,
        endIndex: 1
      },
      properties: {
        pixelSize: 400
      },
      fields: "pixelSize"
    }
  },
  {
    updateDimensionProperties: {
      range: {
        sheetId: runInfoSheetId,
        dimension: "COLUMNS",
        startIndex: 0,
        endIndex: 1
      },
      properties: {
        pixelSize: 200
      },
      fields: "pixelSize"
    }
  },
  {
    repeatCell: {
      range: {
        sheetId,
        startRowIndex: 2,
        startColumnIndex: 0,
        endColumnIndex: 1
      },
      cell: {
        userEnteredFormat: {
          wrapStrategy: "CLIP"
        }
      },
      fields: "userEnteredFormat.wrapStrategy"
    }
  }
];
const getFormattingUpdates = (sheetId, columns) => {
  const batchUpdates = [
    {
      setBasicFilter: {
        filter: {
          range: { sheetId, startRowIndex: 1 }
        }
      }
    },
    {
      repeatCell: {
        range: {
          sheetId,
          startRowIndex: 0,
          endRowIndex: 2
        },
        cell: {
          userEnteredFormat: {
            wrapStrategy: "WRAP",
            horizontalAlignment: "CENTER"
          }
        },
        fields: "userEnteredFormat.wrapStrategy,userEnteredFormat.horizontalAlignment"
      }
    },
    ...columns.flatMap((column, i) => {
      const colIndex = i + 1;
      const colWidth = Math.min(Math.max(column.name.length, column.nameDetail ? column.nameDetail.length + 2 : 0) * 6 + 42, 170);
      const colRange = {
        sheetId,
        startRowIndex: 2,
        startColumnIndex: colIndex,
        endColumnIndex: colIndex + 1
      };
      return [
        ...column.field.type === ColumnType.AuditBoolean ? [
          {
            addConditionalFormatRule: {
              rule: {
                ranges: [colRange],
                booleanRule: {
                  condition: {
                    type: "TEXT_EQ",
                    values: [{ userEnteredValue: "PASS" }]
                  },
                  format: {
                    backgroundColor: colors.good
                  }
                }
              }
            }
          },
          {
            addConditionalFormatRule: {
              rule: {
                ranges: [colRange],
                booleanRule: {
                  condition: {
                    type: "TEXT_EQ",
                    values: [{ userEnteredValue: "FAIL" }]
                  },
                  format: {
                    backgroundColor: colors.bad
                  }
                }
              }
            }
          }
        ] : [
          {
            addConditionalFormatRule: {
              rule: {
                ranges: [colRange],
                gradientRule: column.field.type === ColumnType.AuditValue ? {
                  maxpoint: {
                    colorStyle: { rgbColor: colors.bad },
                    type: "MAX"
                  },
                  midpoint: {
                    colorStyle: { rgbColor: colors.neutral },
                    type: "PERCENT",
                    value: "50"
                  },
                  minpoint: {
                    colorStyle: { rgbColor: colors.good },
                    type: "MIN"
                  }
                } : {
                  maxpoint: {
                    colorStyle: { rgbColor: colors.good },
                    type: "NUMBER",
                    value: "100"
                  },
                  midpoint: {
                    colorStyle: { rgbColor: colors.neutral },
                    type: "PERCENT",
                    value: "50"
                  },
                  minpoint: {
                    colorStyle: { rgbColor: colors.bad },
                    type: "MIN"
                  }
                }
              },
              index: 0
            }
          }
        ],
        {
          updateDimensionProperties: {
            range: {
              sheetId,
              startIndex: colIndex,
              endIndex: colIndex + 1,
              dimension: "COLUMNS"
            },
            properties: {
              pixelSize: colWidth
            },
            fields: "pixelSize"
          }
        },
        {
          repeatCell: {
            range: colRange,
            cell: {
              userEnteredFormat: {
                numberFormat: {
                  type: "NUMBER",
                  pattern: column.field.type === ColumnType.AuditValue ? "#0.0" : "###"
                }
              }
            },
            fields: "userEnteredFormat.numberFormat"
          }
        }
      ];
    })
  ];
  return batchUpdates;
};

const sheetNames = {
  main: "Lighthouse Results",
  auditBreakdown: "Audit Breakdown",
  runInfo: "Run Info"
};
const getDefaultTitle = (initialUrl) => {
  const date = tinydate("{YYYY}-{MM}-{DD} {HH}:{mm}")(new Date());
  return `Lighthouse ${new URL(initialUrl).hostname} ${date}`;
};
const createGoogleSheetsOutputWriter = async (initialUrl, documentTitle = getDefaultTitle(initialUrl)) => {
  let mutexPromise = Promise.resolve();
  const { auth, redirect } = await getAuthenticatedClient();
  const service = googleSheets.sheets({ version: "v4", auth });
  let rowNum = 0;
  const spreadsheet = await service.spreadsheets.create({
    resource: {
      properties: {
        title: documentTitle
      }
    }
  }).then((r) => r);
  const spreadsheetId = spreadsheet.data.spreadsheetId;
  const spreadsheetUrl = spreadsheet.data.spreadsheetUrl;
  console.log(`Output spreadsheet: ${blue(underline(spreadsheetUrl))}`);
  redirect(spreadsheetUrl);
  const updateResponse = await batchUpdate(service, spreadsheetId, [
    {
      addSheet: {
        properties: {
          title: sheetNames.main,
          gridProperties: {
            frozenColumnCount: 1,
            frozenRowCount: 2
          }
        }
      }
    },
    {
      addSheet: {
        properties: { title: sheetNames.auditBreakdown }
      }
    },
    {
      addSheet: {
        properties: { title: sheetNames.runInfo }
      }
    }
  ]);
  const sheetId = updateResponse.data.replies[0].addSheet.properties.sheetId;
  const auditBreakdownSheetId = updateResponse.data.replies[1].addSheet.properties.sheetId;
  const runInfoSheetId = updateResponse.data.replies[2].addSheet.properties.sheetId;
  await batchUpdate(service, spreadsheetId, [
    {
      deleteSheet: {
        sheetId: spreadsheet.data.sheets[0].properties.sheetId
      }
    },
    ...getInitialFormatting(sheetId, runInfoSheetId)
  ]);
  await writeRow(service, spreadsheetId, sheetNames.main, 1, [
    "",
    "Waiting for first Lighthouse report to complete..."
  ]);
  return {
    async writeRunInfo(runInfo) {
      const valuesBatchUpdateRequest = {
        data: [
          {
            range: `'${sheetNames.runInfo}'!A1`,
            values: [
              ["Command", runInfo.command],
              ["Time", runInfo.time],
              [],
              ["node version:", runInfo.versions.node],
              ["npm version:", runInfo.versions.npm],
              ["lighthouse version:", runInfo.versions.lighthouse],
              [
                "lighthouse-parade version:",
                runInfo.versions.lighthouseParade
              ],
              ["chrome version:", runInfo.versions.chrome],
              [],
              ["OS:", runInfo.system.operatingSystem],
              ["Memory:", runInfo.system.memory],
              ["CPU Info", runInfo.system.cpus]
            ]
          }
        ]
      };
      await service.spreadsheets.values.batchUpdate({
        spreadsheetId,
        resource: valuesBatchUpdateRequest,
        valueInputOption: "USER_ENTERED"
      }).then((r) => r);
    },
    async writeHeader(columns) {
      mutexPromise = mutexPromise.then(async () => {
        await writeRow(service, spreadsheetId, sheetNames.main, ++rowNum, [
          "",
          ...columns.map((c) => c.lighthouseCategory)
        ]);
        await writeRow(service, spreadsheetId, sheetNames.main, ++rowNum, [
          "URL",
          ...columns.map((c) => c.name + (c.nameDetail ? `
(${c.nameDetail})` : ""))
        ]);
        const formattingBatchUpdates = getFormattingUpdates(sheetId, columns);
        const {
          batchUpdates: chartBatchUpdates,
          batchValues: chartBatchValues
        } = getCharts(sheetId, auditBreakdownSheetId, columns);
        await batchUpdate(service, spreadsheetId, [
          ...formattingBatchUpdates,
          ...chartBatchUpdates
        ]);
        await service.spreadsheets.values.batchUpdate({
          spreadsheetId,
          resource: chartBatchValues,
          valueInputOption: "USER_ENTERED"
        }).then((r) => r);
      });
      await mutexPromise;
    },
    async addEntry(url, rowValues) {
      mutexPromise = mutexPromise.then(async () => {
        await writeRow(service, spreadsheetId, sheetNames.main, ++rowNum, [
          url,
          ...rowValues
        ]);
      });
      await mutexPromise;
    },
    async complete() {
    }
  };
};
const writeRow = async (service, spreadsheetId, sheetName, rowNum, row) => {
  await service.spreadsheets.values.update({
    spreadsheetId,
    range: `'${sheetName}'!A${rowNum}`,
    valueInputOption: "USER_ENTERED",
    resource: {
      values: [row]
    }
  });
};
const batchUpdate = async (service, spreadsheetId, requests) => {
  const r = await service.spreadsheets.batchUpdate({
    spreadsheetId,
    resource: { requests }
  });
  return r;
};

export { createGoogleSheetsOutputWriter, sheetNames };
